/*
 * gb.c - VXI-11 GPIB library for Agilent E5810A
 *
 * (C) 2008 Kouichi Hirabayashi (kh@mogami-wire.co.jp)
 *
 * Absolutely no warranty. Use this software with your own risk.
 *
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright and disclaimer notice.
 */

#include <stdio.h>
#include <stdlib.h>
#include <rpc/rpc.h>
#include <string.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>

#include "vxi11_xdr.c"
#include "vxi11_clnt.c"

#define LOCK_TIMEOUT 0		/* ms */
#define IO_TIMEOUT   50000	/* ms */

#define	PAD(a)	((a > 99) ? a / 100 : a)

/* Operation Flags */
#define WAITLOCK	0x01		/* waitlock */
#define OF_END		0x08		/* end */
#define TERMCHRSET	0x80		/* termchrset */

static void onint(int);		/* SIGINT */
static void onsrq(int);		/* SIGUSR1 */

static struct sigaction saint, osaint;
static struct sigaction sasrq, osasrq;

typedef struct {
  char *host;		/* XVI-11 server */
  u_long haddr;	/* my IP address */
  struct sockaddr_in addr;	/* clnttcp_create(used in clnttcp_create() */
  int sock;		/* used in clnttcp_create() */
  pid_t c_pid;		/* child (server) pid */
  pid_t p_pid;		/* parent (client) pid */
  CLIENT *clnt_core;	/* Core channel client RPC handle */
  CLIENT *clnt_abort;	/* Abort channel client RPC handle */
  SVCXPRT *svc_tp;	/* server transport handle */
  SVCXPRT *clt_tp;	/* client transport handle */
  u_short srq_port;	/* create_inter_chan RPC port */
  u_long lock_timeout;	/* time to wait for lock */
  u_long io_timeout;	/* time to wait for I/O */
  int reason;		/* latest Device_ReadResp.reason */
  int error;		/* latest RPC error code */
  int srq;		/* SRQ detected */
  int cnt;		/* link counter */
  int ad;		/* current link GPIB address */
  void (*usrfun)(void);	/* user function for SIGINT */
  void (*srqfun)(void);	/* user function for SIGINT */
  int verbose;		/* print some interest informations */
} instr_dev_t;

static instr_dev_t dev;

typedef struct {
  char device[16];	/* GPIB device name (gpib0,23 etc) */
  Device_Link lid;	/* Device Link ID */
  u_short abortPort;	/* Abort RPC port (VXI-11 server) */
  u_long maxRecvSize;	/* max data bytes device will accept on a write */
  int intr;		/* establish interrupt channel (create_intr_chan) */
} instr_link_t;

#define MAX_GPIB_DEV	32
static instr_link_t link[MAX_GPIB_DEV];

static char *errstr[] = {
  "No error", "Syntax error", NULL, "device not accessible",
  "invalid link identifier", "parameter error", "channel not established",
  NULL, "operation not supported", "out of resources", NULL,
  "device locked by another link", "no lock held by this link",
  NULL, NULL, "I/O timeout", NULL, "I/O error", NULL, NULL, NULL,
  "Invalid address", NULL, "abort", NULL, NULL, NULL, NULL, NULL,
  "channel already, established"
};

static char *
strsave(char *s)
{
  char *p;

  if ((p = (char *)malloc(strlen(s) + 1)) != NULL)
	strcpy(p, s);
  else
	p = (char *)NULL;
  return p;
}

int
gbabort(int ad) 
{
  Device_Error *result;

  setLink(ad);
  if ((result = device_abort_1(&link[PAD(ad)].lid, dev.clnt_abort))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_abort, "gbabort()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_abort: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

static int
createLink(int ad)
{
  Create_LinkResp  *result;
  Create_LinkParms  args;
  int pad;
    
  args.lockDevice = FALSE;
  args.lock_timeout= dev.lock_timeout;
  args.device = link[PAD(ad)].device;
  if ((result = create_link_1(&args, dev.clnt_core))
		== (Create_LinkResp *) NULL) {
	clnt_perror(dev.clnt_core, "create_link_1");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("create_klink: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  else {
	if(dev.verbose) {
		printf("create_link deviceLink(%u), abortPort(%hu), MaxRcvSize(%u)\n",
			result->lid, result->abortPort, result->maxRecvSize);
	}
  }
  link[pad = PAD(ad)].lid = result->lid; 
  link[pad].abortPort = result->abortPort;
  link[pad].maxRecvSize = result->maxRecvSize;
  return 0;
}

static int
destroyLink(int ad)
{
  Device_Error *result;

  if ((result = destroy_link_1(&link[PAD(ad)].lid, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "destroy_link_1");
	goto g1;
  }
  link[PAD(ad)].lid = 0; 
  if ((dev.error = result->error) != 0) {
	printf("destroy_link: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

int
gblock(int ad) 
{
  Device_Error *result;
  Device_LockParms args;

  setLink(ad);
  args.lid = link[PAD(ad)].lid;
  args.flags = dev.lock_timeout ? WAITLOCK : 0;
  args.lock_timeout= dev.lock_timeout;

  if ((result = device_lock_1(&args, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "gblock()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_lock: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

int
gbunlock(int ad) 
{
  Device_Error  *result;

  setLink(ad);
  if ((result = device_unlock_1(&link[PAD(ad)].lid, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "gbunlock()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_unlock: (%d) %s\n",
		result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

int
gbwrt(int ad, char *buf, int cnt, int del)
{
  static char *lbuf = (char *)NULL;
  Device_WriteResp *result;
  Device_WriteParms args;
  int d, d0, k, k0, n, max, res;
  char *s, *t, *dlm;

  setLink(ad);
  if (cnt <= 0)
	cnt = strlen(buf);
  if (del > 1) {
	if (lbuf == (char *)NULL)
		lbuf = (char *)malloc(link[PAD(ad)].maxRecvSize);
	t = lbuf;
	dlm = (del > 2) ? "\r" : "\r\n";
  }
  else
	t = buf, d = 0, dlm = "";
  d = d0 = strlen(dlm);
  max = link[PAD(ad)].maxRecvSize;
  for (s = buf, n = 0, res = cnt; n < cnt + d0; ) {
	k = k0 = (res > max) ? max : res;
	if (del > 1) {
		memcpy(t, s, k);
		if (res + d <= max) {
			memcpy(t + k, dlm, d);
			k += d;
			dlm = "";
			d = 0;
		}
		else if (res + 1 <= max) {
			memcpy(t + k, "\r", 1);
			k++;
			if (del == 2) {
				dlm = "\n";
				d = 1;
			}
			else {
				dlm = "";
				d = 0;
			}
		}
	}
	else
		t = s;
	args.lid = link[PAD(ad)].lid;
	args.io_timeout = dev.io_timeout;
	args.lock_timeout = dev.lock_timeout;
	args.flags = 0;
	if (dev.lock_timeout)
		args.flags |= WAITLOCK;
	if (res == k0 && d == 0) {	/* last transfer */
		if (del == 1)
			args.flags |= OF_END;
	}
	args.data.data_val = t;
	args.data.data_len = k;
	if ((result = device_write_1(&args, dev.clnt_core))
			== (Device_WriteResp *) NULL) {
		clnt_perror(dev.clnt_core, "gbwrt()");
		goto g1;
	}
	if ((dev.error = result->error) != 0) {
		printf("device_write: (%d) %s\n", result->error, errstr[result->error]);
g1:
		gbcls();
		exit(1);
	}
	else {
		if(dev.verbose)
			printf("device_write: size= %u\n", result->size);
	}
	n += k;
	s += k0;
	res -= k0;
  }
  return n;
}

int
gbred(int ad, char *buf, int cnt, int del)
{
  Device_ReadResp *result;
  Device_ReadParms args;
  int i, n;

  setLink(ad);
  for (n = 0; n < cnt; ) {
	args.lid = link[PAD(ad)].lid;
	args.requestSize = cnt;
	args.io_timeout = dev.io_timeout;
	args.lock_timeout= dev.lock_timeout;
	args.flags = (del > 1) ? TERMCHRSET : 0x00;
	if (dev.lock_timeout)
		args.flags |= WAITLOCK;
	args.termChar = (del == 3) ? '\r' : '\n';
	if ((result = device_read_1(&args, dev.clnt_core))
			== (Device_ReadResp *) NULL) {
		clnt_perror(dev.clnt_core, "gbred()");
		goto g1;
	}
	dev.reason = result->reason;
	if ((dev.error = result->error) != 0) {
		printf("device_read: (%d) %s\n", result->error, errstr[result->error]);
g1:
		gbcls();
		exit(1);
	}
	memcpy(buf + n, result->data.data_val, result->data.data_len);
	n += result->data.data_len;
	if (result->reason)
		break;	/* END|CHR|REQCNT */
  }
  if (del == 2 && buf[n - 1] == '\n')
	--n;
  if (del > 1 && buf[n - 2] == '\r')
	--n;
  if (n < cnt)
	buf[n] = '\0';
  return n;
}

int
gbspol(int ad)
{
  Device_ReadStbResp *result;
  Device_GenericParms args;
  char *s;

  setLink(ad);
  args.lid = link[PAD(ad)].lid;
  args.io_timeout = dev.io_timeout;
  args.lock_timeout = dev.lock_timeout;
  args.flags = dev.lock_timeout ? WAITLOCK : 0;
  if ((result = device_readstb_1(&args, dev.clnt_core))
		== (Device_ReadStbResp *) NULL) {
	clnt_perror(dev.clnt_core, "gbspol()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_readstb: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return result->stb;
}

int
gbtrg(int ad)
{
  Device_Error *result;
  Device_GenericParms args;

  setLink(ad);
  args.lid  = link[PAD(ad)].lid;
  args.io_timeout = dev.io_timeout;
  args.lock_timeout= dev.lock_timeout;
  args.flags = dev.lock_timeout ? WAITLOCK : 0;
  if ((result = device_trigger_1(&args, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "gbtrg()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_trigger: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

int
gbclr(int ad)
{
  Device_Error *result;
  Device_GenericParms args;

  setLink(ad);
  args.lid = link[PAD(ad)].lid;
  args.io_timeout = dev.io_timeout;
  args.lock_timeout = dev.lock_timeout;
  args.flags = dev.lock_timeout ? WAITLOCK : 0;
  if ((result = device_clear_1(&args, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "gbclr");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_clear: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

int
gbrmt(int ad)
{
  Device_Error *result;
  Device_GenericParms args;

  setLink(ad);
  args.lid = link[PAD(ad)].lid;
  args.io_timeout = dev.io_timeout;
  args.lock_timeout = dev.lock_timeout;
  args.flags = dev.lock_timeout ? WAITLOCK : 0;
  if ((result = device_remote_1(&args, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "gbrmt()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_remote: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

int
gblcl(int ad)
{
  Device_Error *result;
  Device_GenericParms args;

  setLink(ad);
  args.lid = link[PAD(ad)].lid;
  args.io_timeout = dev.io_timeout;
  args.lock_timeout = dev.lock_timeout;
  args.flags = dev.lock_timeout ? WAITLOCK : 0;

  if ((result = device_local_1(&args, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "gblcl()");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_local: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

#if 0	/* not supported by Agilent E5810A */
int
device_Docmd(instr_dev_t *dev, Device_Flags flags, char *cmd)
{
  Device_DocmdResp *result;
  Device_DocmdParms args;

  setLink(ad);
  args.lid = link[PAD(ad)].lid;
  args.io_timeout = dev->io_timeout;
  args.lock_timeout = dev->lock_timeout;
  args.flags = flags;
  args.cmd = 0x020001;	/* Send Command */
  args.network_order = TRUE;
  args.datasize = 1;
  args.data_in.data_in_val = cmd;
  args.data_in.data_in_len = strlen(cmd);
  
  if ((result = device_docmd_1(&args, dev->clnt_core))
		== (Device_DocmdResp *) NULL) {
	clnt_perror (dev->clnt_core, "device_docmd_1");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_docmd: (%d) %s\n", result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
//  printf("*%02x*\n", result->data_out.data_out_val);
  return 0;
}
#endif

static int
createIntrChan(instr_dev_t *dev, Device_AddrFamily progFamily)
{
  Device_Error *result;
  Device_RemoteFunc args;

  if (dev->verbose)
	printf("createIntrChan(): hostAddr(%x), hostPort(%d)\n",
		dev->haddr, dev->srq_port);
  args.hostAddr = htonl(dev->haddr);
  args.hostPort = dev->srq_port;
  args.progNum   = DEVICE_INTR;
  args.progVers  = DEVICE_INTR_VERSION;
  args.progFamily= progFamily;
  if ((result = create_intr_chan_1(&args, dev->clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev->clnt_core, "create_intr_chan_1");
	goto g1;
  }
  if ((dev->error = result->error) != 0) {
	printf("create_intr_chan: (%d) %s\n",
		result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

static int
destroyIntrChan(instr_dev_t *dev)
{
  Device_Error *result;

  if ((result= destroy_intr_chan_1((void*) NULL, dev->clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev->clnt_core, "destroy_intr_chan_1");
	goto g1;
  }
  if ((dev->error = result->error) != 0) {
	printf("destroy_intr_chan: (%d) %s\n",
		result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

static int
deviceEnableSrq(int ad, bool_t enable, u_int handle_len, void *handle_val)
{
  Device_Error *result;
  Device_EnableSrqParms args;

  if (dev.verbose)
	printf("deviceEnableSrq(): lid(%d), enable(%d), handle_len(%d)\n",
		link[PAD(ad)].lid, enable, handle_len);
  args.lid = link[PAD(ad)].lid;
  args.enable = enable;
  args.handle.handle_len = handle_len;
  args.handle.handle_val = handle_val;
  if ((result = device_enable_srq_1(&args, dev.clnt_core))
		== (Device_Error *) NULL) {
	clnt_perror(dev.clnt_core, "device_enable_srq_1");
	goto g1;
  }
  if ((dev.error = result->error) != 0) {
	printf("device_enable_srq: (%d) %s\n",
		result->error, errstr[result->error]);
g1:
	gbcls();
	exit(1);
  }
  return 0;
}

void *
device_intr_srq_1_svc(Device_SrqParms *argp, struct svc_req *rqstp)
{
  static char *result;

  if (dev.verbose)
	fprintf(stderr, "SRQ\n");
  kill(dev.p_pid, SIGUSR1);

  return (void *) &result;
}

void
device_intr_1(struct svc_req *rqstp, register SVCXPRT *transp)
{
  union {
	Device_SrqParms device_intr_srq_1_arg;
  } argument;

  char *result;
  xdrproc_t _xdr_argument, _xdr_result;
  char *(*local)(char *, struct svc_req *);

  switch (rqstp->rq_proc) {
  case NULLPROC:
	(void) svc_sendreply (transp, (xdrproc_t) xdr_void, (char *)NULL);
	return;
  case device_intr_srq:
	_xdr_argument = (xdrproc_t) xdr_Device_SrqParms;
	_xdr_result = (xdrproc_t) xdr_void;
	local = (char *(*)(char *, struct svc_req *)) device_intr_srq_1_svc;
	break;
  default:
	svcerr_noproc (transp);
	return;
  }
  memset ((char *)&argument, 0, sizeof (argument));
  if (!svc_getargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
	svcerr_decode (transp);
	return;
  }
  result = (*local)((char *)&argument, rqstp);
  if (result != NULL && !svc_sendreply(transp, (xdrproc_t) _xdr_result, result)) {
	svcerr_systemerr (transp);
  }
  if (!svc_freeargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
	fprintf (stderr, "%s", "unable to free arguments");
	exit (1);
  }
  return;
}

int
gblost(int time)
{
  dev.io_timeout = time;
  return 0;
}

int
gblost2(int time1, int time2)
{
  dev.io_timeout = time1;
  dev.lock_timeout = time2;
  return 0;
}

int
gbwait(int mask)
{
  int r = 0;
  sigset_t newmask, oldmask;

  if (mask < 0)
	r = dev.reason;		/* END | CHR | REQCNT */
  else if (mask == 0) {
	if (dev.srq) {
		r = 0x40;	/* SRQ */
		sigemptyset(&newmask);
		sigaddset(&newmask, SIGINT);
		if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
			fprintf(stderr, "SIG_BLOCK error.\n");
		dev.srq = 0;
		if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
			fprintf(stderr, "SIG_SETMASK error.\n");
	}
  }
  else {
	while (!dev.srq)
		sleep(600);	/* breaked by SRQ */
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGINT);
	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
		fprintf(stderr, "SIG_BLOCK error.\n");
	dev.srq = 0;
	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
		fprintf(stderr, "SIG_SETMASK error.\n");
	r = 0;
  }
  return r;
}

int
gbopn(char *host)
{
  char host_name[MAXHOSTNAMELEN+1];
  struct hostent *host_ent;
  int i, j;
  int n, status, pfd[2];
  char pbuf[64];

  saint.sa_handler = onint;
  sigemptyset(&saint.sa_mask);
  saint.sa_flags = 0;
  sigaction(SIGINT, &saint, &osaint);

  dev.verbose = 0;
#ifdef USE_SRQ
  gethostname(host_name, sizeof(host_name));
  host_ent= gethostbyname(host_name);
  memcpy(&dev.haddr, host_ent->h_addr_list[0], sizeof(dev.haddr));
  if (dev.verbose)
	printf("Host name= %s address= %hhu.%hhu.%hhu.%hhu\n", host_ent->h_name,
		host_ent->h_addr_list[0][0],
		host_ent->h_addr_list[0][1],
		host_ent->h_addr_list[0][2],
		host_ent->h_addr_list[0][3]);
  sasrq.sa_handler = onsrq;
  sigemptyset(&sasrq.sa_mask);
  sasrq.sa_flags = 0;
  sigaction(SIGUSR1, &sasrq, &osasrq);
  if (pipe(pfd) < 0) {
	printf("pipe error\n");
	exit(1);
  }
  /* SRQ server */
  if ((dev.c_pid = fork()) == 0) {
	register    SVCXPRT *transp;

	close(pfd[0]);
	dev.p_pid = getppid();
	pmap_unset(DEVICE_INTR, DEVICE_INTR_VERSION);
	transp = svctcp_create(RPC_ANYSOCK, 0, 0);
	if (transp == NULL) {
		fprintf (stderr, "%s", "cannot create tcp service.\n");
		exit(1);
	}
	if (!svc_register(transp, DEVICE_INTR, DEVICE_INTR_VERSION, device_intr_1, IPPROTO_TCP)) {
		fprintf (stderr, "%s", "unable to register (DEVICE_INTR, DEVICE_INTR_VERSION, tcp).\n");
		exit(1);
	}
	dev.svc_tp = transp;
	if (dev.verbose)
		printf("Interrupt (SRQ) port(%hu)\n", transp->xp_port);
	sprintf(pbuf, "%d", transp->xp_port);
	if (write(pfd[1], pbuf, strlen(pbuf)) != strlen(pbuf)) {
		printf("write() error.\n");
	}
	close(pfd[1]);
	svc_run();
	fprintf(stderr, "svc_run() returned !\n");
	exit(1);
  }
  /* Core Channel */
  close(pfd[1]);
  if ((n = read(pfd[0], pbuf, sizeof(pbuf))) <= 0) {
	printf("read error.\n");
	exit(1);
  }
  pbuf[n] = '\0';
  close(pfd[0]);
  sscanf(pbuf, "%d", &dev.srq_port);
  if (dev.verbose)
	printf("srq_port: %s(%d)\n", pbuf, dev.srq_port);
#endif
  if (dev.verbose)
	printf("VXI-11 server: %s\n", host);

  host_ent= gethostbyname(host);
  if (dev.verbose)
	printf("Host name= %s address= %hhu.%hhu.%hhu.%hhu\n", host_ent->h_name,
		host_ent->h_addr_list[0][0],
		host_ent->h_addr_list[0][1],
		host_ent->h_addr_list[0][2],
		host_ent->h_addr_list[0][3]);
  dev.host = strsave(host_ent->h_name);
  bzero((char *)&dev.addr, sizeof(dev.addr));
  dev.addr.sin_addr.s_addr = *(u_long *)host_ent->h_addr_list[0];
  dev.addr.sin_family = AF_INET;
  dev.addr.sin_port = 0;
  dev.sock = RPC_ANYSOCK;
  dev.clnt_core= clnttcp_create(&dev.addr, DEVICE_CORE, DEVICE_CORE_VERSION,
	&dev.sock, 0, 0);
  if (dev.clnt_core == NULL) {
	fprintf(stderr, "Core: ");
	kill(dev.c_pid, SIGINT);
	clnt_pcreateerror (host);
	exit (1);
  }
  dev.lock_timeout = LOCK_TIMEOUT;
  dev.io_timeout = IO_TIMEOUT;
  return 0;
}

int
setLink(int ad)
{
  int pad;	/* primary address */

  pad = PAD(ad);
  if (link[pad & 0x3F].lid == (Device_Link)NULL) {	/* create link */
	if (pad < 0 || 31 < pad || (ad % 100) > 31) {
		fprintf(stderr, "Illegale GPIB address %d.\n", ad);
		exit(1);
	}
	dev.ad = ad;
	if (ad == 31)	/* 31 for COM1 */
		sprintf(link[pad].device, "COM1");
	else if (ad < 100)
		sprintf(link[pad].device, "gpib%d,%d", 0, ad);
	else
		sprintf(link[pad].device, "gpib%d,%d,%d", 0, ad / 100, ad % 100);
	createLink(pad);
	if (dev.verbose)
		printf("Abort_port(%u), Max send message size(%d)\n",
			link[ad].abortPort, link[ad].maxRecvSize);
	/* Abort Channel */
	dev.addr.sin_port = link[pad].abortPort;
	dev.clnt_abort = clnttcp_create(&dev.addr, DEVICE_ASYNC, DEVICE_ASYNC_VERSION,
		&dev.sock, 0, 0);
	if (dev.clnt_abort == NULL) {
		fprintf(stderr, "Abort: ");
		clnt_pcreateerror(dev.host);
		exit (1);
	}

#ifdef USE_SRQ
	if (!dev.cnt) {
		createIntrChan(&dev, DEVICE_TCP);
		deviceEnableSrq(ad, TRUE, sizeof(void*),(void*) &dev);
		link[PAD(ad)].intr = 1;
	}
#endif
	dev.cnt++;
  }
  return 0;
}

int
gbcls(void)
{
  int i;

  for (i = 0; i < 31; i++) {
#ifdef USE_SRQ
	if (link[i].intr != (Device_Link)NULL)
		deviceEnableSrq(i, FALSE, sizeof(void*),(void*) &dev);
#endif
	if (link[i].lid != (Device_Link)NULL)
		destroyLink(i);
  }
#ifdef USE_SRQ
  if (dev.svc_tp != (SVCXPRT *)NULL)
	destroyIntrChan(&dev);
#endif
  if (dev.clnt_core != NULL)
	clnt_destroy(dev.clnt_core);
  if (dev.svc_tp != NULL)
	clnt_destroy(dev.clnt_abort);
  kill(dev.c_pid, SIGTERM);
  pmap_unset(DEVICE_INTR, DEVICE_INTR_VERSION);

  return 0;
}

int
gberr()
{
  return dev.error;
}

static void
onint(int sig)
{
  if (!dev.p_pid) {
	gbcls();
	if (dev.usrfun != (void *)NULL)
		(*dev.usrfun)();
  }
  else {
	svc_unregister(DEVICE_INTR, DEVICE_INTR_VERSION);
  }
  exit(0);
}

static void
onsrq(int sig)
{
  sasrq.sa_handler = onsrq;
  sigemptyset(&sasrq.sa_mask);
  sasrq.sa_flags = 0;
  sigaction(SIGUSR1, &sasrq, &osasrq);
  dev.srq = 1;
  if (dev.srqfun != (void *)NULL)
		(*dev.srqfun)();
  if (dev.verbose)
	printf("SIGUSR1\n");
}

void
gbint(void *fun)
{
  dev.usrfun = fun;
}

void
gbsrq(void *fun)
{
  dev.srqfun = fun;
}

/* INTEX 207X compatible functions */

int
gbcmod(int mode)
{
  if (mode != 3)
	fprintf(stderr, "gbcmod(%d) is not supported.\n", mode);
  return 0;
}

int
gbifc(void)
{
  int i;

  for (i = 0; i < MAX_GPIB_DEV; i++)
	if (link[i].lid != (Device_Link)NULL)
		break;
  if (i >= MAX_GPIB_DEV)
	return -1;
  return gbabort(i);
}

int
gbren(int f)
{
  int i;

  for (i = 0; i < MAX_GPIB_DEV; i++)
	if (link[i].lid != (Device_Link)NULL)
		break;
  if (i >= MAX_GPIB_DEV)
	return -1;
  if (f)
	return gbrmt(i);
  return gblcl(i);
}

int
gblst(int time)
{
  dev.io_timeout = time * 1000;
  return 0;
}

int
issrq(void)
{
  return gbwait(0);
}

int
gbadr(int ad)
{
  fprintf(stderr, "gbadr() is not supported.\n");
  return -1;
}

int
gbmsg(char *msg)
{
  fprintf(stderr, "gbmsg() is not supported.\n");
  return -1;
}

int
gbppol(int ad)
{
  fprintf(stderr, "gbppol() is not supported.");
  return -1;
}

int
gbppc(int ad, int data)
{
  fprintf(stderr, "gbppc() is not supported.\n");
  return -1;
}

int
gbppu(int ad)
{
  fprintf(stderr, "gbppu() is not supported.\n");
  return -1;
}
